var N3Store = require('../N3').Store;
var vows = require('vows'),
    chai = require('chai'),
    util = require('util');
chai.should();
chai.use(require('chai-things'));

vows.describe('N3Store').addBatch({
  'The N3Store module': {
    topic: function () { return N3Store; },

    'should be a function': function (N3Store) {
      N3Store.should.be.a('function');
    },

    'should make N3Store objects': function (N3Store) {
      N3Store().should.be.an.instanceof(N3Store);
    },

    'should be an N3Store constructor': function (N3Store) {
      new N3Store().should.be.an.instanceof(N3Store);
    },
  },

  'An empty N3Store': {
    topic: new N3Store(),

    'should have size 0': function (n3Store) {
      n3Store.size.should.eql(0);
    },

    'should be empty': function (n3Store) {
      n3Store.find().should.be.empty;
    },

    'should have a default context': function (n3Store) {
      n3Store.defaultContext.should.eql('n3/contexts#default');
    },

    'should be able to create unnamed blank nodes': function (n3Store) {
      n3Store.createBlankNode().should.eql('_:b0');
      n3Store.createBlankNode().should.eql('_:b1');

      n3Store.add('_:b0', '_:b1', '_:b2');
      n3Store.createBlankNode().should.eql('_:b3');
    },

    'should be able to create named blank nodes': function (n3Store) {
      n3Store.createBlankNode('blank').should.eql('_:blank');
      n3Store.createBlankNode('blank').should.eql('_:blank1');
      n3Store.createBlankNode('blank').should.eql('_:blank2');
    },
  },

  'An N3Store with initialized with 3 elements': {
    topic: function () {
      var n3Store = new N3Store([
        { subject: 's1', predicate: 'p1', object: 'o1'},
        { subject: 's1', predicate: 'p1', object: 'o2'},
        { subject: 's1', predicate: 'p1', object: 'o3'},
      ]);
      return n3Store;
    },

    'should have size 3': function (n3Store) {
      n3Store.size.should.eql(3);
    },
  },

  'An N3Store with 5 elements': {
    topic: function () {
      var n3Store = new N3Store();
      n3Store.add('s1', 'p1', 'o1');
      n3Store.addTriple({ subject: 's1', predicate: 'p1', object: 'o2'});
      n3Store.addTriples([
        { subject: 's1', predicate: 'p2', object: 'o2'},
        { subject: 's2', predicate: 'p1', object: 'o1'},
      ]);
      n3Store.add('s1', 'p2', 'o3', 'c4');
      return n3Store;
    },

    'should have size 5': function (n3Store) {
      n3Store.size.should.eql(5);
    },

    'when searched without parameters': {
      topic: function (n3Store) { return n3Store.find(); },

      'should return all items in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s1', 'p1', 'o2'], ['s1', 'p2', 'o2'], ['s2', 'p1', 'o1'])
    },

    'when searched with an existing subject parameter': {
      topic: function (n3Store) { return n3Store.find('s1', null, null); },

      'should return all items with this subject in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s1', 'p1', 'o2'], ['s1', 'p2', 'o2'])
    },

    'when searched with a non-existing subject parameter': {
      topic: function (n3Store) { return n3Store.find('s3', null, null); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with an existing predicate parameter': {
      topic: function (n3Store) { return n3Store.find(null, 'p1', null); },

      'should return all items with this predicate in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s1', 'p1', 'o2'], ['s2', 'p1', 'o1'])
    },

    'when searched with a non-existing predicate parameter': {
      topic: function (n3Store) { return n3Store.find(null, 'p3', null); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with an existing object parameter': {
      topic: function (n3Store) { return n3Store.find(null, null, 'o1'); },

      'should return all items with this object in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s2', 'p1', 'o1'])
    },

    'when searched with a non-existing object parameter': {
      topic: function (n3Store) { return n3Store.find(null, null, 'o4'); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with existing subject and predicate parameters': {
      topic: function (n3Store) { return n3Store.find('s1', 'p1', null); },

      'should return all items with this subject and predicate in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s1', 'p1', 'o2'])
    },

    'when searched with non-existing subject and predicate parameters': {
      topic: function (n3Store) { return n3Store.find('s2', 'p2', null); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with existing subject and object parameters': {
      topic: function (n3Store) { return n3Store.find('s1', null, 'o2'); },

      'should return all items with this subject and object in the default context':
        shouldIncludeAll(['s1', 'p1', 'o2'], ['s1', 'p2', 'o2'])
    },

    'when searched with non-existing subject and object parameters': {
      topic: function (n3Store) { return n3Store.find('s2', 'p2', null); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with existing predicate and object parameters': {
      topic: function (n3Store) { return n3Store.find(null, 'p1', 'o1'); },

      'should return all items with this predicate and object in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s2', 'p1', 'o1'])
    },

    'when searched with non-existing predicate and object parameters': {
      topic: function (n3Store) { return n3Store.find(null, 'p2', 'o3'); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with existing subject, predicate, and object parameters': {
      topic: function (n3Store) { return n3Store.find('s1', 'p1', 'o1'); },

      'should return all items with this subject, predicate, and object in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'])
    },

    'when searched with non-existing subject, predicate, and object parameters': {
      topic: function (n3Store) { return n3Store.find('s2', 'p2', 'o2'); },

      'should return no items': shouldBeEmpty()
    },

    'when searched with the default context parameter': {
      topic: function (n3Store) { return n3Store.find(); },

      'should return all items in the default context':
        shouldIncludeAll(['s1', 'p1', 'o1'], ['s1', 'p1', 'o2'], ['s1', 'p2', 'o2'], ['s2', 'p1', 'o1'])
    },

    'when searched with an existing non-default context parameter': {
      topic: function (n3Store) { return n3Store.find(null, null, null, 'c4'); },

      'should return all items in that context':
        shouldIncludeAll(['s1', 'p2', 'o3', 'c4'])
    },

    'when searched with a non-existing non-default context parameter': {
      topic: function (n3Store) { return n3Store.find(null, null, null, 'c5'); },

      'should return no items': shouldBeEmpty()
    },
  },

  'An N3Store': {
    topic: new N3Store(),

    // Test inspired by http://www.devthought.com/2012/01/18/an-object-is-not-a-hash/.
    // The value `__proto__` is not supported however – fixing it introduces too much overhead.
    'should be able to contain entities with JavaScript object property names': function (n3Store) {
      n3Store.add('toString', 'valueOf', 'toLocaleString', 'hasOwnProperty');
      shouldIncludeAll(['toString', 'valueOf', 'toLocaleString', 'hasOwnProperty'])
        (n3Store.find(null, null, null, 'hasOwnProperty'));
    },

    'should be able to contain entities named "null"': function (n3Store) {
      n3Store.add('null', 'null', 'null', 'null');
      shouldIncludeAll(['null', 'null', 'null', 'null'])
        (n3Store.find(null, null, null, 'null'));
    },
  },
}).export(module);

function shouldBeEmpty() {
  return function (result) {
    result.should.be.empty;
  };
}

function shouldIncludeAll() {
  var items = Array.prototype.map.call(arguments, function (arg) {
    return { subject: arg[0], predicate: arg[1], object: arg[2],
             context: arg[3] || 'n3/contexts#default' };
  });
  return function (result) {
    result.should.have.length(items.length);
    for (var i = 0; i < items.length; i++)
      result.should.include.something.that.deep.equals(items[i]);
  };
}
